r"""پرتوهای نور که توسط لنزها (Lens) شکست داده می‌شوند."""

from __future__ import annotations
from typing import Iterable


import numpy as np
from manim import *
from shapely import geometry as gm

from manim_fa_physics.rays import intersection
from manim_fa_physics.mechanics import snell, antisnell, rotate_vector, angle_of_vector

from manim import config, Line
from manim.utils.space_ops import angle_of_vector, rotate_vector
import numpy as np


__all__ = ["Ray"]


class Ray(Line):
    def __init__(
        self,
        start: Iterable[float],
        direction: Iterable[float],
        init_length: float = 5,
        propagate: Iterable[Lens] | None = None,
        **kwargs,
    ) -> None:
        """یک پرتو نور.

        پارامترها
        ----------
        start
            نقطه شروع پرتو
        direction
            جهت پرتو
        init_length
            طول اولیه پرتو. پس از عبور از لنزها، طول پرتو برای نمایش شکست تغییر می‌کند.
        propagate
            لیستی از لنزها که پرتو باید از آن‌ها عبور کند.

        مثال
        -------
        .. manim:: RayExampleScene
            :save_last_frame:

            from manim_physics import *

            class RayExampleScene(Scene):
                def construct(self):
                    lens_style = {"fill_opacity": 0.5, "color": BLUE}
                    a = Lens(-5, 1, **lens_style).shift(LEFT)
                    a2 = Lens(5, 1, **lens_style).shift(RIGHT)
                    b = [
                        Ray(LEFT * 5 + UP * i, RIGHT, 8, [a, a2], color=RED)
                        for i in np.linspace(-2, 2, 10)
                    ]
                    self.add(a, a2, *b)
        """
        self.init_length = init_length
        self.propagated = False
        super().__init__(start, np.array(start) + np.array(direction) * init_length, **kwargs)
        if propagate:
            self.propagate(*propagate)

    def propagate(self, *lenses: Lens) -> None:
        """عبور پرتو از لیست لنزهای داده شده.

        پارامترها
        ----------
        lenses
            تمام لنزهایی که پرتو باید از آن‌ها عبور کند.
        """
        sorted_lens = self._sort_lens(lenses)
        for lens in sorted_lens:
            intersects = intersection(lens, self)
            if len(intersects) == 0:
                continue
            intersects = self._sort_intersections(intersects)
            if not self.propagated:
                self.put_start_and_end_on(self.start, intersects[1])
            else:
                nppcc = self.n_points_per_cubic_curve if config.renderer != "opengl" else self.n_points_per_curve
                self.points = self.points[:-nppcc]
                self.add_line_to(intersects[1])
            self.end = intersects[1]

            # زاویه تابش و شکست
            i_ang = angle_of_vector(self.end - lens.C[0])
            i_ang -= angle_of_vector(self.start - self.end)
            r_ang = snell(i_ang, lens.n)
            r_ang *= -1 if lens.f > 0 else 1

            ref_ray = rotate_vector(lens.C[0] - self.end, r_ang)
            intersects = intersection(
                lens,
                Line(self.end - ref_ray * self.init_length, self.end + ref_ray * self.init_length)
            )
            intersects = self._sort_intersections(intersects)
            self.add_line_to(intersects[1])

            self.start = self.end
            self.end = intersects[1]

            # برخورد با سطح دوم لنز
            i_ang = angle_of_vector(self.end - lens.C[1])
            i_ang -= angle_of_vector(self.start - self.end)
            if np.abs(np.sin(i_ang)) < 1 / lens.n:
                r_ang = antisnell(i_ang, lens.n)
                r_ang *= -1 if lens.f < 0 else 1
                ref_ray = rotate_vector(lens.C[1] - self.end, r_ang)
                ref_ray *= -1 if lens.f > 0 else 1
                self.add_line_to(self.end + ref_ray * self.init_length)
                self.start = self.end
                self.end = self.get_end()

            self.propagated = True

    def _sort_lens(self, lenses: Iterable[Lens]) -> Iterable[Lens]:
        """مرتب‌سازی لنزها بر اساس فاصله از نقطه شروع پرتو"""
        dists = []
        for lens in lenses:
            try:
                dists += [[np.linalg.norm(intersection(self, lens)[0] - self.start), lens]]
            except:
                dists += [[np.inf, lens]]
        dists.sort(key=lambda x: x[0])
        return np.array(dists, dtype=object)[:, 1]

    def _sort_intersections(self, intersections: Iterable[Iterable[float]]) -> Iterable[Iterable[float]]:
        """مرتب‌سازی نقاط تقاطع بر اساس فاصله تا انتهای پرتو"""
        result = []
        for inters in intersections:
            result.append([np.linalg.norm(inters - self.end), inters])
        result.sort(key=lambda x: x[0])
        return np.array(result, dtype=object)[:, 1]



r"""پاندول‌ها.

کلاس‌های :class:`~MultiPendulum` و :class:`~Pendulum` هر دو از ویژگی
:py:mod:`~rigid_mechanics` منشعب شده‌اند.

"""

from __future__ import annotations
from typing import Iterable, Optional
from functools import partial

from manim import VGroup, Mobject, Line, Circle, UP, DOWN, RIGHT, np
from manim.utils.color import ORANGE
import pymunk

from .mechanics import SpaceScene

__all__ = [
    "Pendulum",
    "MultiPendulum",
    "SpaceScene",
]


class MultiPendulum(VGroup):
    def __init__(
        self,
        *bobs: Iterable[np.ndarray],
        pivot_point: np.ndarray = UP * 2,
        rod_style: Optional[dict] = None,
        bob_style: Optional[dict] = None,
        **kwargs,
    ) -> None:
        """یک چندپاندول.

        پارامترها
        ----------
        bobs
            موقعیت توپ‌های پاندول.
        pivot_point
            نقطه اتکا یا محور پاندول.
        rod_style
            پارامترهای مربوط به خطوط اتصال.
        bob_style
            پارامترهای مربوط به توپ‌ها (Circle).
        kwargs
            پارامترهای اضافی برای VGroup.

        مثال
        -------
        .. manim:: MultiPendulumExample
            :quality: low

            from manim_physics import *

            class MultiPendulumExample(SpaceScene):
                def construct(self):
                    p = MultiPendulum(RIGHT, LEFT)
                    self.add(p)
                    self.make_rigid_body(*p.bobs)
                    p.start_swinging()
                    self.add(TracedPath(p.bobs[-1].get_center, stroke_color=BLUE))
                    self.wait(10)
        """
        if rod_style is None:
            rod_style = {}
        if bob_style is None:
            bob_style = {"radius": 0.1, "color": ORANGE, "fill_opacity": 1}

        self.pivot_point = pivot_point
        self.bobs = VGroup(*[Circle(**bob_style).move_to(i) for i in bobs])
        self.pins = [pivot_point] + list(bobs)
        self.rods = VGroup()
        self.rods.add(Line(self.pivot_point, self.bobs[0].get_center(), **rod_style))
        self.rods.add(
            *(
                Line(self.bobs[i].get_center(), self.bobs[i + 1].get_center(), **rod_style)
                for i in range(len(bobs) - 1)
            )
        )

        super().__init__(**kwargs)
        self.add(self.rods, self.bobs)

    def _make_joints(
        self, mob1: Mobject, mob2: Mobject, spacescene: SpaceScene
    ) -> None:
        a = mob1.body
        if isinstance(mob2, np.ndarray):
            b = pymunk.Body(body_type=pymunk.Body.STATIC)
            b.position = mob2[0], mob2[1]
        else:
            b = mob2.body
        joint = pymunk.PinJoint(a, b)
        spacescene.space.space.add(joint)

    def _redraw_rods(self, mob: Line, pins, i):
        try:
            x, y, _ = pins[i]
        except:
            x, y = pins[i].body.position
        x1, y1 = pins[i + 1].body.position
        mob.put_start_and_end_on(
            np.array([x, y, 0]),
            np.array([x1, y1, 0]),
        )

    def start_swinging(self) -> None:
        """شروع حرکت پاندول‌ها."""
        spacescene: SpaceScene = self.bobs[0].spacescene
        pins = [self.pivot_point] + list(self.bobs)

        for i in range(len(pins) - 1):
            self._make_joints(pins[i + 1], pins[i], spacescene)
            # استفاده از partial برای جلوگیری از ارور lambda در حلقه
            self.rods[i].add_updater(partial(self._redraw_rods, pins=pins, i=i))

    def end_swinging(self) -> None:
        """توقف حرکت پاندول‌ها."""
        spacescene = self.bobs[0].spacescene
        spacescene.stop_rigidity(self.bobs)


class Pendulum(MultiPendulum):
    def __init__(
        self,
        length: float = 3.5,
        initial_theta: float = 0.3,
        pivot_point: np.ndarray = UP * 2,
        rod_style: Optional[dict] = None,
        bob_style: Optional[dict] = None,
        **kwargs,
    ):
        """یک پاندول ساده.

        پارامترها
        ----------
        length
            طول پاندول.
        initial_theta
            زاویه اولیه انحراف از محور قائم.
        pivot_point
            نقطه اتکا.
        rod_style
            پارامترهای خطوط اتصال.
        bob_style
            پارامترهای توپ پاندول.
        kwargs
            پارامترهای اضافی VGroup.

        مثال
        -------
        .. manim:: PendulumExample
            :quality: low

            from manim_physics import *
            class PendulumExample(SpaceScene):
                def construct(self):
                    pends = VGroup(*[Pendulum(i) for i in np.linspace(1, 5, 7)])
                    self.add(pends)
                    for p in pends:
                        self.make_rigid_body(*p.bobs)
                        p.start_swinging()
                    self.wait(10)
        """
        if rod_style is None:
            rod_style = {}
        if bob_style is None:
            bob_style = {"radius": 0.25, "color": ORANGE, "fill_opacity": 1}

        self.length = length
        self.pivot_point = pivot_point

        point = self.pivot_point + (
            RIGHT * np.sin(initial_theta) * length
            + DOWN * np.cos(initial_theta) * length
        )
        super().__init__(
            point,
            pivot_point=self.pivot_point,
            rod_style=rod_style,
            bob_style=bob_style,
            **kwargs,
        )



r"""ماژول الکترواستاتیک (Electrostatics)"""

from __future__ import annotations
from typing import Iterable

from manim import normalize, VGroup, Dot, Rectangle, Arc
from manim.constants import ORIGIN, TAU
from manim.mobject.vector_field import ArrowVectorField
from manim.utils.color import BLUE, RED, RED_A, RED_D, color_gradient
import numpy as np

__all__ = ["Charge", "ElectricField"]


class Charge(VGroup):
    """یک جسم بار الکترواستاتیک که یک :class:`~ElectricField` تولید می‌کند."""

    def __init__(
        self,
        magnitude: float = 1,
        point: np.ndarray = ORIGIN,
        add_glow: bool = True,
        **kwargs,
    ) -> None:
        """پارامترها
        ----------
        magnitude
            شدت بار الکترواستاتیک
        point
            موقعیت بار
        add_glow
            آیا افکت درخشندگی اضافه شود. حلقه‌های متنوعی برای شبیه‌سازی نور ایجاد می‌کند.
        kwargs
            پارامترهای اضافی برای VGroup
        """
        VGroup.__init__(self, **kwargs)
        self.magnitude = magnitude
        self.point = point
        self.radius = (abs(magnitude) * 0.4 if abs(magnitude) < 2 else 0.8) * 0.3

        if magnitude > 0:
            label = VGroup(
                Rectangle(width=0.32 * 1.1, height=0.006 * 1.1).set_z_index(1),
                Rectangle(width=0.006 * 1.1, height=0.32 * 1.1).set_z_index(1),
            )
            color = RED
            layer_colors = [RED_D, RED_A]
            layer_radius = 4
        else:
            label = Rectangle(width=0.27, height=0.003)
            color = BLUE
            layer_colors = ["#3399FF", "#66B2FF"]
            layer_radius = 2

        if add_glow:  # استفاده از چندین Arc برای شبیه‌سازی درخشندگی
            layer_num = 80
            color_list = color_gradient(layer_colors, layer_num)
            opacity_func = lambda t: 1500 * (1 - abs(t - 0.009) ** 0.0001)
            rate_func = lambda t: t**2

            for i in range(layer_num):
                self.add(
                    Arc(
                        radius=layer_radius * rate_func((0.5 + i) / layer_num),
                        angle=TAU,
                        color=color_list[i],
                        stroke_width=101 * (rate_func((i + 1) / layer_num) - rate_func(i / layer_num)) * layer_radius,
                        stroke_opacity=opacity_func(rate_func(i / layer_num)),
                    ).shift(point)
                )

        self.add(Dot(point=self.point, radius=self.radius, color=color))
        self.add(label.scale(self.radius / 0.3).shift(point))
        for mob in self:
            mob.set_z_index(1)


class ElectricField(ArrowVectorField):
    """یک میدان الکتریکی."""

    def __init__(self, *charges: Charge, **kwargs) -> None:
        """پارامترها
        ----------
        charges
            تمام بارهایی که میدان الکتریکی را ایجاد می‌کنند
        kwargs
            پارامترهای اضافی برای ArrowVectorField
        """
        self.charges = charges
        positions = []
        magnitudes = []
        for charge in charges:
            positions.append(charge.get_center())
            magnitudes.append(charge.magnitude)

        super().__init__(lambda p: self._field_func(p, positions, magnitudes), **kwargs)

    def _field_func(
        self,
        p: np.ndarray,
        positions: Iterable[np.ndarray],
        magnitudes: Iterable[float],
    ) -> np.ndarray:
        """محاسبه بردار میدان الکتریکی در نقطه p"""
        field_vect = np.zeros(3)
        for p0, mag in zip(positions, magnitudes):
            r = p - p0
            dist = np.linalg.norm(r)
            if dist < 0.1:  # جلوگیری از تقسیم بر صفر
                return np.zeros(3)
            field_vect += mag / dist**2 * normalize(r)
        return field_vect



r"""ماژول مغناطیس ایستا (Magnetostatics)"""

from __future__ import annotations
import itertools as it
from typing import Iterable, Tuple

from manim import VMobject, ArrowVectorField, np
from manim.mobject.opengl.opengl_compatibility import ConvertToOpenGL

__all__ = ["Wire", "MagneticField"]


class Wire(VMobject, metaclass=ConvertToOpenGL):
    """کلاس انتزاعی برای نمایش یک سیم حامل جریان
    که یک :class:`~MagneticField` تولید می‌کند.

    پارامترها
    ----------
    stroke
        VMobject اصلی سیم. سیم خروجی فرم آن را می‌گیرد.
    current
        مقدار جریان عبوری از سیم.
    samples
        تعداد قطعات سیم برای ایجاد :class:`~MagneticField`.
    kwargs
        پارامترهای اضافی برای VMobject.

    .. note::
        برای مثال، به :class:`~MagneticField` مراجعه کنید.
    """

    def __init__(
        self,
        stroke: VMobject,
        current: float = 1,
        samples: int = 16,
        **kwargs,
    ):
        self.current = current
        self.samples = samples
        super().__init__(**kwargs)
        self.set_points(stroke.points)


class MagneticField(ArrowVectorField):
    """یک میدان مغناطیسی.

    پارامترها
    ----------
    wires
        تمام سیم‌هایی که در تولید میدان نقش دارند.
    kwargs
        پارامترهای اضافی برای ArrowVectorField.

    مثال
    -------
    .. manim:: MagneticFieldExample
        :save_last_frame:

        from manim_physics import *

        class MagneticFieldExample(ThreeDScene):
            def construct(self):
                wire = Wire(Circle(2).rotate(PI / 2, UP))
                mag_field = MagneticField(
                    wire,
                    x_range=[-4, 4],
                    y_range=[-4, 4],
                )
                self.set_camera_orientation(PI / 3, PI / 4)
                self.add(wire, mag_field)
    """

    def __init__(self, *wires: Wire, **kwargs):
        dls = []
        currents = []
        for wire in wires:
            points = [
                wire.point_from_proportion(i)
                for i in np.linspace(0, 1, wire.samples + 1)
            ]
            dls.append(list(zip(points, points[1:])))
            currents.append(wire.current)

        super().__init__(
            lambda p: MagneticField._field_func(p, dls, currents),
            **kwargs
        )

    @staticmethod
    def _field_func(
        p: np.ndarray,
        dls: Iterable[Tuple[np.ndarray, np.ndarray]],
        currents: Iterable[float],
    ):
        """محاسبه بردار میدان مغناطیسی در نقطه p"""
        B_field = np.zeros(3)
        for dl in dls:
            for (r0, r1), I in it.product(dl, currents):
                dr = r1 - r0
                r = p - r0
                dist = np.linalg.norm(r)
                if dist < 0.1:  # جلوگیری از تقسیم بر صفر
                    return np.zeros(3)
                B_field += np.cross(dr, r) * I / dist**4
        return B_field



r"""لنزها برای شکست پرتوها (Rays)."""

from __future__ import annotations
from typing import Iterable, Tuple

from manim import config, VMobject, VectorizedPoint, Circle, Square
from manim import Difference

from manim.utils.color import WHITE
import numpy as np
from shapely import geometry as gm

__all__ = ["Lens"]

try:
    # برای manim < 0.15.0
    from manim.mobject.opengl_compatibility import ConvertToOpenGL
except ModuleNotFoundError:
    # برای manim >= 0.15.0
    from manim.mobject.opengl.opengl_compatibility import ConvertToOpenGL


def intersection(vmob1: VMobject, vmob2: VMobject) -> np.ndarray:
    """نقاط تقاطع دو منحنی"""
    a = gm.LineString(vmob1.points)
    b = gm.LineString(vmob2.points)
    intersects: gm.GeometryCollection = a.intersection(b)
    try:  # اگر چندین تقاطع وجود داشته باشد
        return np.array(
            [[[x, y, z] for x, y, z in m.coords][0] for m in intersects.geoms]
        )
    except:  # در غیر این صورت
        return np.array([[x, y, z] for x, y, z in intersects.coords])


def snell(i_ang: float, n: float) -> float:
    """قانون اسنل برای شکست نور
    ورودی و خروجی بر حسب رادیان
    """
    return np.arcsin(np.sin(i_ang) / n)


def antisnell(r_ang: float, n: float) -> float:
    """قانون معکوس اسنل
    ورودی و خروجی بر حسب رادیان
    """
    return np.arcsin(np.sin(r_ang) * n)


class Lens(VMobject, metaclass=ConvertToOpenGL):
    def __init__(self, f: float, d: float, n: float = 1.52, **kwargs) -> None:
        """یک لنز. معمولاً همراه با :class:`~Ray` استفاده می‌شود.

        پارامترها
        ----------
        f
            فاصله کانونی. با مقدار مثبت لنز محدب، با مقدار منفی لنز مقعر.
            توجه: مقدار f دقیقاً با نقطه کانونی همخوانی ندارد (مشکل شناخته شده).
        d
            ضخامت لنز
        n
            ضریب شکست. پیش‌فرض برای شیشه.
        kwargs
            پارامترهای اضافی برای VMobject
        """
        super().__init__(**kwargs)
        self.f = f
        f *= 50 / 7 * f if f > 0 else -50 / 7 * f  # روش غیرمعمول اما کار می‌کند
        if f > 0:
            r = ((n - 1) ** 2 * f * d / n) ** 0.5
        else:
            r = ((n - 1) ** 2 * -f * d / n) ** 0.5
        self.d = d
        self.n = n
        self.r = r

        # ساخت هندسه لنز
        if f > 0:
            self.set_points(
                Intersection(
                    a := Circle(r).shift(np.array([r - d / 2, 0, 0])),
                    b := Circle(r).shift(np.array([-(r - d / 2), 0, 0])),
                )
                .insert_n_curves(50)
                .points
            )
        else:
            self.set_points(
                Difference(
                    Difference(
                        Square(2 * 0.7 * r),
                        a := Circle(r).shift(np.array([-(r + d / 2), 0, 0])),
                    ),
                    b := Circle(r).shift(np.array([r + d / 2, 0, 0])),
                )
                .insert_n_curves(50)
                .points
            )

        # افزودن نقاط مرکزی هندسه برای محاسبات بعدی
        self.add(VectorizedPoint(a.get_center()), VectorizedPoint(b.get_center()))

    @property
    def C(self) -> Tuple[np.ndarray, np.ndarray]:
        """بازگرداندن دو نقطه متناظر با مراکز انحنای لنز"""
        i = 0
        i += 1 if config.renderer != "opengl" else 0
        return self[i].points[0], self[i + 1].points[0]  # این کمی گیج‌کننده است



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setup(
    name="manim-fa-physics",
    version="0.1.0",
    author="Tabesh Alli",
    description="افزونه فیزیک برای Manim (مکانیک، گرانش، اوپتیک، الکترومغناطیس، الکترواستاتیک)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Tabesh2020/manim-fa-physics",
    packages=find_packages(),
    include_package_data=True,  # ← اضافه شد
    install_requires=["manim>=0.19.0", "numpy>=1.24"],
    python_requires=">=3.10",
    license="MIT",  # ← اختیاری ولی توصیه‌شده
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Multimedia :: Graphics",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from .mechanics import *
from .pendulum import *
from .lenses import *
from .rays import *
from .electrostatics import *
from .magnetostatics import *

__all__ = [
    *mechanics.__all__,
    *pendulum.__all__,
    *lenses.__all__,
    *rays.__all__,
    *electrostatics.__all__,
    *magnetostatics.__all__,
]



r"""فضای شبیه‌سازی گرانش.

بیشتر اشیاء می‌توانند به صورت یک جسم صلب (حرکت طبق گرانش و برخورد)
یا یک جسم ساکن (ثابت در صحنه) تعریف شوند.

برای استفاده از این ویژگی، باید از :class:`~SpaceScene` استفاده شود
تا به توابع مخصوص فضای شبیه‌سازی دسترسی داشته باشید.

.. note::
    * این ویژگی از بسته pymunk استفاده می‌کند. اگرچه ضروری نیست،
      اما آشنایی مختصر با نحوه استفاده از آن مفید است.

      `مستندات رسمی <http://www.pymunk.org/en/latest/pymunk.html>`_

      `آموزش یوتیوب <https://youtu.be/pRk---rdrbo>`_

    * نرخ فریم پایین ممکن است باعث شود برخی اشیاء از اشیاء ساکن عبور کنند،
      زیرا برخوردها به اندازه کافی دقیق ثبت نمی‌شوند. افزایش نرخ فریم
      ممکن است این مشکل را حل کند.

مثال
--------
.. manim:: TwoObjectsFalling

    from manim_physics import *
    # استفاده از SpaceScene برای دسترسی به تمام متدهای مکانیک صلب
    class TwoObjectsFalling(SpaceScene):
        def construct(self):
            circle = Circle().shift(UP)
            circle.set_fill(RED, 1)
            circle.shift(DOWN + RIGHT)

            rect = Rectangle().shift(UP)
            rect.rotate(PI / 4)
            rect.set_fill(YELLOW_A, 1)
            rect.shift(UP * 2)
            rect.scale(0.5)

            ground = Line([-4, -3.5, 0], [4, -3.5, 0])
            wall1 = Line([-4, -3.5, 0], [-4, 3.5, 0])
            wall2 = Line([4, -3.5, 0], [4, 3.5, 0])
            walls = VGroup(ground, wall1, wall2)
            self.add(walls)

            self.play(
                DrawBorderThenFill(circle),
                DrawBorderThenFill(rect),
            )
            self.make_rigid_body(rect, circle)  # اشیاء با گرانش حرکت می‌کنند
            self.make_static_body(walls)        # اشیاء در محل ثابت می‌مانند
            self.wait(5)
"""
from __future__ import annotations
import numpy as np
from manim import *
from typing import Tuple

from manim import VGroup, VMobject, Mobject, Circle, Line, Rectangle, Polygon, Polygram, Scene, np
from manim.utils.space_ops import angle_between_vectors
import pymunk

__all__ = [
    "Space",
    "_step",
    "_simulate",
    "get_shape",
    "get_angle",
    "SpaceScene",
]


class Space(Mobject):
    def __init__(self, gravity: Tuple[float, float] = (0, -9.81), **kwargs):
        """یک شی انتزاعی برای گرانش.

        پارامترها
        ----------
        gravity
            جهت و شدت گرانش.
        """
        super().__init__(**kwargs)
        self.space = pymunk.Space()
        self.space.gravity = gravity
        self.space.sleep_time_threshold = 5


class SpaceScene(Scene):
    GRAVITY: Tuple[float, float] = (0, -9.81)

    def __init__(self, renderer=None, **kwargs):
        """صحنه پایه برای تمام اشیاء مکانیک صلب.
        وکتور گرانش با ``self.GRAVITY`` قابل تنظیم است.
        """
        self.space = Space(gravity=self.GRAVITY)
        super().__init__(renderer=renderer, **kwargs)

    def setup(self):
        """استفاده داخلی"""
        self.add(self.space)
        self.space.add_updater(_step)

    def add_body(self, body: Mobject):
        """اضافه کردن جسم به pymunk و اتصال Mobject به آن."""
        if body.body != self.space.space.static_body:
            self.space.space.add(body.body)
        self.space.space.add(body.shape)

    def make_rigid_body(
        self,
        *mobs: Mobject,
        elasticity: float = 0.8,
        density: float = 1,
        friction: float = 0.8,
    ):
        """تبدیل هر Mobject به جسم قابل حرکت توسط گرانش.

        پارامترها
        ----------
        mobs
            اشیاء برای تبدیل به جسم صلب.
        elasticity, density, friction
            ویژگی‌های برخورد و تعامل با اشیاء دیگر.
        """
        for mob in mobs:
            if not hasattr(mob, "body"):
                self.add(mob)
                mob.body = pymunk.Body()
                mob.body.position = mob.get_x(), mob.get_y()
                get_angle(mob)
                if not hasattr(mob, "angle"):
                    mob.angle = 0
                mob.body.angle = mob.angle
                get_shape(mob)
                mob.shape.density = density
                mob.shape.elasticity = elasticity
                mob.shape.friction = friction
                mob.spacescene = self

                self.add_body(mob)
                mob.add_updater(_simulate)
            else:
                if mob.body.is_sleeping:
                    mob.body.activate()

    def make_static_body(
        self, *mobs: Mobject, elasticity: float = 1, friction: float = 0.8
    ) -> None:
        """تبدیل هر Mobject به جسم ساکن برای تعامل با اشیاء صلب."""
        for mob in mobs:
            if isinstance(mob, (VGroup, list)):
                self.make_static_body(*mob)
                continue
            mob.body = self.space.space.static_body
            get_shape(mob)
            mob.shape.elasticity = elasticity
            mob.shape.friction = friction
            self.add_body(mob)

    def stop_rigidity(self, *mobs: Mobject) -> None:
        """توقف حرکت صلب اشیاء."""
        for mob in mobs:
            if isinstance(mob, (VGroup, list)):
                self.stop_rigidity(*mob)
                continue
            if hasattr(mob, "body"):
                mob.body.sleep()


def _step(space, dt):
    """به‌روزرسانی گام فیزیکی"""
    space.space.step(dt)


def _simulate(b):
    """به‌روزرسانی موقعیت و زاویه Mobject طبق pymunk Body"""
    x, y = b.body.position
    b.move_to(np.array([x, y, 0]))
    b.rotate(b.body.angle - getattr(b, "angle", 0))
    b.angle = b.body.angle


def get_shape(mob: VMobject) -> None:
    """دریافت شکل جسم از Mobject"""
    if isinstance(mob, Circle):
        mob.shape = pymunk.Circle(body=mob.body, radius=mob.radius)
    elif isinstance(mob, Line):
        mob.shape = pymunk.Segment(
            mob.body,
            (mob.get_start()[0], mob.get_start()[1]),
            (mob.get_end()[0], mob.get_end()[1]),
            mob.stroke_width - 3.95,
        )
    elif isinstance(mob, Rectangle):
        width = np.linalg.norm(mob.get_vertices()[1] - mob.get_vertices()[0])
        height = np.linalg.norm(mob.get_vertices()[2] - mob.get_vertices()[1])
        mob.shape = pymunk.Poly.create_box(mob.body, (width, height))
    elif isinstance(mob, Polygram):
        vertices = [(a, b) for a, b, _ in mob.get_vertices() - mob.get_center()]
        mob.shape = pymunk.Poly(mob.body, vertices)
    elif isinstance(mob, Polygon):
        width = np.linalg.norm(mob.get_vertices()[1] - mob.get_vertices()[0])
        height = np.linalg.norm(mob.get_vertices()[2] - mob.get_vertices()[1])
        mob.shape = pymunk.Poly.create_box(mob.body, (width, height))
    else:
        mob.shape = pymunk.Poly.create_box(mob.body, (mob.width, mob.height))


def get_angle(mob: VMobject) -> None:
    """دریافت زاویه جسم از Mobject، استفاده داخلی برای Updaterها"""
    if isinstance(mob, Polygon):
        vec1 = mob.get_vertices()[0] - mob.get_vertices()[1]
        vec2 = type(mob)().get_vertices()[0] - type(mob)().get_vertices()[1]
        mob.angle = angle_between_vectors(vec1, vec2)
    elif isinstance(mob, Line):
        mob.angle = mob.get_angle()
def snell(i_ang, n):
    """
    قانون اسنل: زاویه شکست را بر اساس زاویه تابش و ضریب شکست لنز محاسبه می‌کند.
    i_ang : زاویه تابش (به رادیان)
    n : ضریب شکست
    """
    # اگر مقدار خیلی نزدیک صفر باشد (برای جلوگیری از خطای ریاضی)
    if abs(np.sin(i_ang)) < 1e-6:
        return 0
    try:
        return np.arcsin(np.sin(i_ang) / n)
    except ValueError:
        # اگر سینوس از محدوده مجاز خارج شد
        return 0


def antisnell(i_ang, n):
    """
    حالت معکوس قانون اسنل (از لنز به هوا)
    """
    if abs(np.sin(i_ang)) < 1e-6:
        return 0
    try:
        return np.arcsin(np.sin(i_ang) * n)
    except ValueError:
        return 0